! Convenience class for dealing with Courant-Snyder (or Twiss) parameters.
! These parameters represent an ellipse in phase space given by
!
!      gamma*x^2 + 2*alpha*x*x' + beta*x'^2 = emitt
!
! Recall that these parameters are related by the fact
!
!      beta*gamma - alpha^2 = 1
!
!
!	Author	: Christopher K. Allen
!	Date	: November, 2005
!

BeginPackage[TwissUtility`];

!------------------------------
!
! Function GetEnvelopeFromTwiss
!
!------------------------------
!
! Get the values of the beam envelope parameters from the corresponding Twiss 
! parameters.  The units depend upon the units of the Twiss parameters (e.g.,
! if the Twiss parameters are in MKS than so are the envelope parameters).
! If the emittance is given as effective emittance (i.e. 5*RMS for 3D beam, 
! 4*RMS for 2D beam), then the envelope parameters are that of the equivalent
! uniform beam.  If the emittance is given as RMS emittance, then the returned 
! values are for the RMS envelope.
!
!    Parameters
!      vecTwiss    vectors {alpha,beta, emittance} of Twiss parameters
!
!    Returned Value
!                  vector of beam envelope parameters {radius,slope,emittance}
!

GetEnvelopeFromTwiss[{alpha_, beta_, emittance_}] := With[{
  radius =  Sqrt[emittance * beta],
  slope  = -Sqrt[emittance / beta] * alpha},

  {radius, slope, emittance}];

!------------------------------
!
! Function GetTwissFromEnvelope
!
!------------------------------
!
! Get the values of the Twiss parameters from the corresponding beam envelope 
! parameters. The units depend upon the units of the envelope parameters (e.g.,
! if the envelope parameters are in MKS than so are the Twiss parameters).
! If the beam envelope parameters are for the RMS envelope, then the returned
! Twiss parameters are the RMS values. If the beam envelope parameters are for
! the equivalent uniform beam then so are the return Twiss parameters.
!
!    Parameters
!      vecTwiss    vectors {radius,slope,emittance} of beam envelope parameters
!
!    Returned Value
!                  vector of Twiss parameters {alpha,beta,emittance}
!

GetTwissFromEnvelope[{radius_, slope_, emittance_}] := With[{
  alpha = -radius * slope  / emittance,
  beta  =  radius * radius / emittance},

  {alpha, beta, emittance}];


!------------------------------
!
! Function CorrelationMatrix
!
!------------------------------
!
! Compute and return the 2x2 correlation matrix for the Twiss parameters.
! The correlation matrix sigma has the form
!
!        sigma = <(x x')*(x x')'>
!              = emitt * |  beta  -alpha |
!                        | -alpha  gamma |
!
!    Parameters
!      vecTwiss    vector {alpha,beta, emittance} of Twiss parameters
!
!    Returned Value
!                  2x2 beam correlation matrix
!

CorrelationMatrix[{alpha_, beta_, emittance_}] := With[{
  gamma = (1 + alpha * alpha) / beta},

  { {  beta * emittance, -alpha * emittance},
    {-alpha * emittance,  gamma * emittance}}];


!--------------------------------
!
! Function CorrelationMatrixTrans
!
!--------------------------------
! 
! Compute and return the 4x4 correlation matrix for the given transverse Twiss 
! parameters. This matrix is composed of the two 2x2 blocks for each transverse
! phase plane. Thus, the returned matrix is the uncoupled correlation matrix.
! sigma and has the form
!
!        sigma = | sigX    0  |
!                |   0   sigY |
!
!    Parameters
!      vecTwissX   vector {alpha,beta, emittance} of X Plane Twiss parameters
!      vecTwissY   vector {alpha,beta, emittance} of Y Plane Twiss parameters
!
!    Returned Value
!                  4x4 beam correlation matrix
!

CorrelationMatrixTrans[vecTwissX_, vecTwissY_] := With[{
  matSigX = CorrelationMatrix[vecTwissX],	! X plane correlation matrix
  matSigY = CorrelationMatrix[vecTwissY],	! Y plane correlation matrix
  matZero = {{0, 0}, {0, 0}}},			! 2x2 zero matrix

  Join[
    MapThread[Join, {matSigX, matZero}],
    MapThread[Join, {matZero, matSigY}],
    Null[]]];


!--------------------------------
!
! Function CorrelationMatrix6D
!
!--------------------------------
!
! Compute and return the full 6x6 correlation matrix for the given the Twiss 
! parameters for all three phase planes.  This matrix is composed of three 
! 2x2 blocks for each phase plane, including the longitudinal.
! The returned matrix is the uncoupled correlation matrix sigma and has the form
!
!        sigma = | sigX    0     0  |
!                |   0   sigY    0  |
!                |   0     0   sigZ | 
!
!    Parameters
!      vecTwissX   vector {alpha,beta, emittance} of X Plane Twiss parameters
!      vecTwissY   vector {alpha,beta, emittance} of Y Plane Twiss parameters
!      vecTwissZ   vector {alpha,beta, emittance} of Z Plane Twiss parameters
!
!    Returned Value
!                  6x6 beam correlation matrix
!

CorrelationMatrix6D[vecTwissX_, vecTwissY_, vecTwissZ_] := With[{
  matSigX = CorrelationMatrix[vecTwissX],	! X plane correlation matrix
  matSigY = CorrelationMatrix[vecTwissY],	! Y plane correlation matrix
  matSigZ = CorrelationMatrix[vecTwissZ],	! Z plane correlation matrix
  matZero = {{0, 0}, {0, 0}}},			! 2x2 zero matrix

  Join[
    MapThread[Join, {matSigX, matZero, matZero}],
    MapThread[Join, {matZero, matSigY, matZero}],
    MapThread[Join, {matZero, matZero, matSigZ}],
    Null[]]];


Protect[GetEnvelopeFromTwiss, GetTwissFromEnvelope];
Protect[CorrelationMatrix, CorrelationMatrixTrans, CorrelationMatrix6D];

EndPackage[];
