! Frontend part for synchronous EPICS CA interface
!
! CaOpen/CaRead/CaWrite/CaGet/CaPut/CaClose APIs are originally developed by N. Yamamoto since 1996(Packages/CaSad.n)
! Ca* APIs are maintained by K. Oide since 1997.
! CA connection manager(CaManager class) is introduced by K. Oide since 1999.
! CA* API implementation is overhauled for new EPICS CA backend at 2017.12 by A. Morita.
!
Unprotect[CaOpen, CaClose, CaRead, CaWrite, CaGet, CaPut];

Class;

BeginPackage[CAManager`];

Begin[CAManager`];

With[{Modifiers = String},

CAManager = Class[{},
  { Handler = True, PreferTkSense = False},
    
  {},

  (* Public interfaces *)

! CaOpen[] interface
  CaOpen[pv1_, pvs__] := OpenPV[{pv1, pvs}];
  CaOpen[pv_] := OpenPV[pv];
  CaOpen[] := $Failed;

! CaClose[] interface
  CaClose[ch1_, chs__] := ClosePV[{ch1, chs}];
  CaClose[ch_] := ClosePV[ch];
  CaClose[] := Null;

! CaRead[] interface
  CaRead[ch1_, chs__] := With[{rl = ExpandCaReadList[{ch1, chs}]},
    With[{chl = OpenPV[First[rl]], convert = Last[rl]},
      convert@@($CaRead2@@chl)]];
  CaRead[ch_] := Switch[ch,
    _List,	With[{rl = ExpandCaReadList[ch]},
      With[{chl = OpenPV[First[rl]], convert = Last[rl]},
	convert@@($CaRead2@@chl)]],
    _,		First[$CaRead2[OpenPV[ch]]],
    _,		$Failed];
  CaRead[___] := $Failed;

! CaWrite[] interface
  CaWrite[pvl_List, val_] := With[{wl = ExpandCaWriteList2[pvl, val]},
    With[{chl = OpenPV[wl[[1, , 1]]], vl = wl[[1, , 2]], convert = wl[[2]]},
      CaWrite2@@Thread[{FilterChannelList[chl], vl}];
      convert@@($CaRead2@@chl)]];
  CaWrite[pv:Rule[Modifiers[_String], _], val_] := CaWrite[OpenPV[pv], val];
  CaWrite[ch:Rule[Modifiers[_Real],   _], val_] := (CaWrite2[{ch[[1, 1]], val}]; First[$CaRead2[ch]]);
  CaWrite[Rule[(String|Real)[___], _], val_] := $Failed;
  CaWrite[pv:Rule[_String,         _], val_] := CaWrite[OpenPV[pv], val];
  CaWrite[ch:Rule[_Real,           _], val_] := (CaWrite2[{ch[[1]],    val}]; First[$CaRead2[ch]]);
  CaWrite[pv:Modifiers[_String], val_] := CaWrite[OpenPV[pv], val];
  CaWrite[ch:Modifiers[_Real],   val_] := (CaWrite2[{ch[[1]], val}]; First[$CaRead2[ch]]);
  CaWrite[(String|Real)[___], val_] := $Failed;
  CaWrite[pv_String,          val_] := CaWrite[OpenPV[pv], val];
  CaWrite[ch_Real,            val_] := (CaWrite2[{ch,      val}]; First[$CaRead2[ch]]);
  CaWrite[wl$_List] := With[{wl = ExpandCaWriteList1[wl$]},
    With[{chl = OpenPV[wl[[1, , 1]]], vl = wl[[1, , 2]], convert = wl[[2]]},
      CaWrite2@@Thread[{FilterChannelList[chl], vl}];
      convert@@($CaRead2@@chl)]];
  CaWrite[___] := $Failed;

  (* Internal cache & helper implementation *)
  (* Opened PV cache *)
  Channel[_] := $Failed;

  (* Blocking CaOpen: List of PVname -> List of chid/$Failed *)
  OpenPV$[pvl_List] := If[Handler, OpenPV2[pvl], OpenPV1[pvl]];
  OpenPV$[pv_String] := First[If[Handler, OpenPV2[{pv}], OpenPV1[{pv}]]];

  (* OpenPV1: ca_create_channel w/o connection handler[CaOpen1] *)
  OpenPV1[pvl_List] := Check[CaOpen1[pvl], Table[$Failed, {Length[pvl]}]];

  (* OpenPV1: ca_create_channel w   connection handler[EPICS$CaOpen] *)
  OpenPV2[pvl_List] := If[pvl === {}, {}, With[{dT = 5e-3,
    timeout = EPICS$CaOpenTimeout[], chl = Check[EPICS$CaOpen[pvl], Table[$Failed, {Length[pvl]}]]},
    If[Not[Or@@RealQ/@chl], chl, (
      EPICS$CaFlushIO[];	(* Flush ca_create_channel requests at first *)
      If[timeout == 0,
	While[True,
	  $SleepAndPoll[dT];
	  If[And@@(CaConnectedQ/@chl), Break[]]],
	Do[
	  $SleepAndPoll[dT];
	  If[And@@(CaConnectedQ/@chl), Break[]],
	  {timeout / dT}]];
      Map[If[CaConnectedQ[#], #, EPICS$CaClearChannel[#]; $Failed]&, chl])]]];

  (* Open PV with Channel cache *)
  OpenPV[pvl_List, cache_Real:False] := If[cache, OpenPV[#, True]&/@pvl,
    With[{pvl$ = Map[Switch[#,
      Rule[Modifiers[_String], _],	If[RealQ[Channel[#[[1, 1]]]], Null[], #[[1, 1]]],
      Rule[String[___], _],		Null[],
      Rule[_String, _],			If[RealQ[Channel[#[[1]]]],    Null[], #[[1]]],
      Modifiers[_String],		If[RealQ[Channel[#[[1]]]],    Null[], #[[1]]],
      String[___],			Null[],
      _String,				If[RealQ[Channel[#]],         Null[], #],
      _,				Null[]]&, Flatten[pvl]]},
      ScanThread[If[RealQ[#2], Channel[#1] = #2]&, {pvl$, OpenPV$[pvl$]}];
      OpenPV[#, True]&/@pvl]];
  OpenPV[Rule[pv:Modifiers[_String], count_], cache_Real:False] := With[{ch = OpenPV[pv, cache]},
    If[ch <=> $Failed, Rule[ch, count], $Failed]];
  OpenPV[Rule[ch:Modifiers[_Real],   count_], cache_Real:False] := Rule[ch, count];
  OpenPV[Rule[(String|Real)[___],    count_], cache_Real:False] := $Failed;
  OpenPV[Rule[pv_String,             count_], cache_Real:False] := With[{ch = OpenPV[pv, cache]},
    If[ch <=> $Failed, Rule[ch, count], $Failed]];
  OpenPV[Rule[ch_Real,               count_], cache_Real:False] := Rule[ch, count];
  OpenPV[pv$:Modifiers[_String], cache_Real:False] := With[{mod = Head[pv$], pv = First[pv$]}, Which[
    RealQ[Channel[pv]],	mod[Channel[pv]],
    cache,		$Failed,
    True,		With[{ch = OpenPV$[pv]}, If[RealQ[ch], (Channel[pv] = ch; mod[ch]), ch]]]];
  OpenPV[ch:Modifiers[_Real],    cache_Real:False] := ch;
  OpenPV[(String|Real)[___],     cache_Real:False] := $Failed;
  OpenPV[pv_String,              cache_Real:False] := If[cache || RealQ[Channel[pv]], Channel[pv],
    With[{ch = OpenPV$[pv]}, If[RealQ[ch], Channel[pv] = ch]; ch]];
  OpenPV[ch_Real,                cache_Real:False] := ch;
  OpenPV[_, cache_Real:False] := $Failed;

  (* Close PV on Channel cache *)
  ClosePV[l_List] := With[{pair = Thread[Map[Switch[#,
    Rule[Modifiers[_String], _],	With[{pv = #[[1, 1]]},
      If[RealQ[Channel[pv]], {pv, Channel[pv]}, Null[]]],
    Rule[Modifiers[_Real],   _],	With[{ch = #[[1, 1]], pv = Check[CaChName[#[[1, 1]]], Null]},
      If[Channel[pv] === ch, {pv, ch}, Null[]]],
    Rule[(String|Real)[___], _],	Null[],
    Rule[_String,            _],	With[{pv = #[[1]]},
      If[RealQ[Channel[pv]], {pv, Channel[pv]}, Null[]]],
    Rule[_Real,              _],	With[{ch = #[[1]],    pv = Check[CaChName[#[[1]]],    Null]},
      If[Channel[pv] === ch, {pv, ch}, Null[]]],
    Modifiers[_String],			With[{pv = #[[1]]},
      If[RealQ[Channel[pv]], {pv, Channel[pv]}, Null[]]],
    Modifiers[_Real],			With[{ch = #[[1]],    pv = Check[CaChName[#[[1]]],    Null]},
      If[Channel[pv] === ch, {pv, ch}, Null[]]],
    (String|Real)[___],	Null[],
    _String,				With[{pv = #},
      If[RealQ[Channel[pv]], {pv, Channel[pv]}, Null[]]],
    _Real,				With[{ch = #,         pv = Check[CaChName[#],         Null]},
      If[Channel[pv] === ch, {pv, ch}, Null[]]],
    _,					Null[]]&, Flatten[l]]]},
    If[pair <=> {},
      Scan[(Channel[#] =.)&, pair[[1]]];
      CaClose1@@pair[[2]]]];
  ClosePV[Rule[Modifiers[pv_String], _]] := With[{ch = Channel[pv]},
    If[RealQ[ch],          Channel[pv] =.; CaClose1[ch]]];
  ClosePV[Rule[Modifiers[ch_Real],   _]] := With[{pv = Check[CaChName[ch], Null]},
    If[Channel[pv] === ch, Channel[pv] =.; CaClose1[ch]]];
  ClosePV[Rule[(String|Real)[___],   _]] := Null;
  ClosePV[Rule[pv_String,            _]] := With[{ch = Channel[pv]},
    If[RealQ[ch],          Channel[pv] =.; CaClose1[ch]]];
  ClosePV[Rule[ch_Real,              _]] := With[{pv = Check[CaChName[ch], Null]},
    If[Channel[pv] === ch, Channel[pv] =.; CaClose1[ch]]];
  ClosePV[Modifiers[pv_String]] := With[{ch = Channel[pv]},
    If[RealQ[ch],          Channel[pv] =.; CaClose1[ch]]];
  ClosePV[Modifiers[ch_Real]]   := With[{pv = Check[CaChName[ch], Null]},
    If[Channel[pv] === ch, Channel[pv] =.; CaClose1[ch]]];
  ClosePV[(String|Real)[___]]   := Null;
  ClosePV[pv_String]            := With[{ch = Channel[pv]},
    If[RealQ[ch],          Channel[pv] =.; CaClose1[ch]]];
  ClosePV[ch_Real]              := With[{pv = Check[CaChName[ch], Null]},
    If[Channel[pv] === ch, Channel[pv] =.; CaClose1[ch]]];

  ClosePV[_] := Null;
  
  (* CaRead helper *)
  ExpandCaReadList[chl_List, index$_Real:-1] := Module[{index = Max[0, index$], l},
    l = Map[Switch[#,
      _List,		With[{sub = ExpandCaReadList[#, index]}, index = First[sub]; Rest[sub]],
      Rule[Modifiers[_String|_Real], _Real],	{$Slot[++index], #},
      Rule[(String|Real)[___],       _Real],	{$Failed, {}},
      Rule[_String|_Real,            _Real],	{$Slot[++index], #},
      Modifiers[_String|_Real],			{$Slot[++index], #},
      (String|Real)[___],			{$Failed, {}},
      _String|_Real,				{$Slot[++index], #},
      _,					{$Failed, {}}]&, chl];
    If[index$ < 0,
      { Flatten[l[[, 2]]],
	With[{expr = l[[, 1]]},	Function[expr]/.$Slot->Slot]},
      {index, l[[, 1]], Flatten[l[[, 2]]]}]];

  (* CaWrite helper *)
  ExpandCaWriteList1[wl_List, index$_Real:-1] := Module[{index = Max[0, index$], l},
    l = Map[Switch[#,
      _[Rule[Modifiers[_String|_Real], _Real], _],	{$Slot[++index], Rule@@#},
      _[Rule[(String|Real)[___],       _Real], _],	{$Failed, {}},
      _[Rule[_String|_Real,            _Real], _],	{$Slot[++index], Rule@@#},
      _[Modifiers[_String|_Real], _],			{$Slot[++index], Rule@@#},
      _[(String|Real)[___],       _],			{$Failed, {}},
      _[_String|_Real,            _],			{$Slot[++index], Rule@@#},
      _[_List, _],	With[{sub = ExpandCaWriteList2[#[[1]], #[[2]], index]},
	index = First[sub]; Rest[sub]],
      _[_List],		With[{sub = ExpandCaWriteList1[#[[1]], index]},
	index = First[sub]; Rest[sub]],
      _,		{$Failed, {}}]&, wl];
    If[index$ < 0,
      { Flatten[l[[, 2]]],
	With[{expr = l[[, 1]]},	Function[expr]/.$Slot->Slot]},
      {index, l[[, 1]], Flatten[l[[, 2]]]}]];

  ExpandCaWriteList2[chl_List, val_, index$_Real:-1] := Module[{index = Max[0, index$], l},
    l = MapThread[Switch[#1,
      _List,		With[{sub = ExpandCaWriteList2[#1, #2, index]}, index = First[sub]; Rest[sub]],
      Rule[Modifiers[_String|_Real], _],	{$Slot[++index], Rule[#1, #2]},
      Rule[(String|Real)[___],    _],		{$Failed, {}},
      Rule[_String|_Real,         _],		{$Slot[++index], Rule[#1, #2]},
      Modifiers[_String|_Real],			{$Slot[++index], Rule[#1, #2]},
      (String|Real)[___],			{$Failed, {}},
      _String|_Real,				{$Slot[++index], Rule[#1, #2]},
      _,					{$Failed, {}}]&, {chl, val}];
    If[index$ < 0,
      { Flatten[l[[, 2]]],
	With[{expr = l[[, 1]]},	Function[expr]/.$Slot->Slot]},
      {index, l[[, 1]], Flatten[l[[, 2]]]}]];

  FilterChannelList[chl_List] := Map[Switch[#,
    Rule[_[_Real], _],		#[[1, 1]],	(* Rule[modifier[chid_Real], count_] *)
    Rule[Real[___], _],		$Failed,
    Rule[_Real, _],		#[[1]],		(* Rule[chid_Real, count_] *)
    _[_Real],			#[[1]],		(* modifier[chid_Real] *)
    Real[___],			$Failed,
    _Real,			#,		(* chid_Real *)
    _,				$Failed]&, chl];
  
  (* CaRead2 exception wrapper *)
  $CaRead2[argv__] := With[{r = If[Handler, CaRead2CB[argv], CaRead2IO[argv]]},
    Switch[Unevaluated$[r],
      _List,	r,
      _,	Table[$Failed, {Length[{argv}]}]]];

  $SleepAndPoll[dt_Real] := If[PreferTkSense && TkinterQ[],
    TkSense[dt$], EPICS$CaPendEvent[dt]];

  ];
];

Protect[CAManager];

End[];

EndPackage[];

(* Export public interfaces into Global scope *)
With[{`scope = CAManager`CAManager},

! CaOpen[] interface
!  Setup CA connection by PV name and return channel ID: chid(successful) or $Failed(failure).
!
!  CaOpen[chid_Real]   return chid
!  CaOpen[pv_String]   return chid	if connection successful
!                             $Failed	otherwise
!  CaOpen[pv_List]     return Map[CaOpen, pv] by efficient way
!  CaOpen[_]           retuen $Failed
!  CaOpen[pv1_, pvs__] return Map[CaOpen, {pv1, pvs}] by efficient way
!  CaOpen[]            return $Failed
!
  CaOpen[argv___] := `scope@CaOpen[argv];

! CaClose[] interface
!  Destruct CA connection made by CaOpen[] interface and return Null.
!
!  CaClose[chid_Real]       destruct CA connection(chid) created by CaOpen[]
!  CaClose[pv_String]       destruct CA connection created by CaOpen[pv]
!  CaClose[chid_List]       do Scan[CaClose, chid] by efficient way
!  CaClose[_]               return Null
!  CaClose[chid1_, chids__] do Scan[CaClose, {chid1, chids}] by efficient way
!  CaClose[]                return Null
!
  CaClose[argv___] := `scope@CaClose[argv];

! CaRead[] interface
!  Read PV via CA connection and return {value, alarm-status, alarm-severity, time-stamp}(successful) or Null(timeout).
!  CAUTION: Internal errors are notified via Message[].
!
!  CaRead[chid_Real]             return PV value via CA connection(chid)
!  CaRead[chid_Real->count_Real] return PV value up-to `count' elements pointer via CA connection(chid)
!  CaRead[pv_String]             return CaRead[CaOpen[pv]]
!  CaRead[pv_String->count_Real] return CaRead[CaOpen[pv]->count]
!  CaRead[rl_List]               return Map[CaRead, rl] by efficient way
!  CaRead[chid1_, chids__]       return Map[CaRead, {chid1, chids}] by efficient way
!
  CaRead[argv___] := `scope@CaRead[argv];

! CaWrite[] interface
!  Write PV via CA connection and return CaRead[] result.
!  CAUTION: Internal errors are notified via Message[].
!
!  CaWrite[chid_Real, value_]    send `value' via CA connection(chid) and return CaRead[chid] result
!  CaWrite[pv_String, value_]    do CaWrite[CaOpen[pv], value]
!  CaWrite[chl_List, vl_List]    do MapThread[CaWrite, {chl, vl}] by efficient way
!  CaWrite[wl_List]              do Apply[CaWrite, wl, {1}] by efficient way
!
  CaWrite[argv___] := `scope@CaWrite[argv];

  CaGet[ch_] := CaRead[ch];
  CaPut[ch_, val_]:=CaWrite[ch, val];

  ];

Protect[CaOpen, CaClose, CaRead, CaWrite, CaGet, CaPut];
